// This is the main DLL file.
#pragma warning(disable:4996) // fopen deprecated

#include <windows.h>
#include <string>
#include <stdio.h>
#include "out_shppoints.h"
#include "shp.h"

using namespace System::Windows::Forms;
using namespace std;


void ExportPointsToDBF(string filename, byte* image, int stride, int width, int height, int count)
{
	string             file_dbf = filename + ".dbf";
	FILE              *fp  = fopen(file_dbf.c_str(), "wb");
	unsigned char      data1;
	unsigned short int data2;
	unsigned int       data4;
	char               field[] = "TYPE\0\0\0\0\0\0\0N\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
	char               semmi[] = "                    ";
	int                fieldlen = 4;
	int                color_index;

	if (fp == 0) return;

	fwrite(&(data1 = 3),               1,  1, fp);  // dBase5
	fwrite(&(data1 = 2010 - 1900),     1,  1, fp);  // year
	fwrite(&(data1 = 5),               1,  1, fp);  // month
	fwrite(&(data1 = 27),              1,  1, fp);  // day
	fwrite(&(data4 = count),           4,  1, fp);  // number of records
	fwrite(&(data2 = 32 + 1 * 32 + 1), 2,  1, fp);  // header length
	fwrite(&(data2 = fieldlen),        2,  1, fp);  // record length
	fwrite(&(data4 = 0),               4,  1, fp);  // 0
	fwrite(&data4,                     4,  1, fp);  // 0
	fwrite(&data4,                     4,  1, fp);  // 0
	fwrite(&data4,                     4,  1, fp);  // 0
	fwrite(&data4,                     4,  1, fp);  // 0

	// field descriptor array
	field[16] = fieldlen - 1; // field length
	fwrite(field,                      1, 32, fp);  // field descriptor
	fwrite(&(data1 = 0x0D),            1,  1, fp);  // header terminator

	for (int y = 0; y < height; y++)
	{
		for (int x = 0; x < width; x++)
		{
			if (color_index = *image++)
			{
				int leading_spaces = fieldlen - 3;
				if (color_index < 100) leading_spaces++;
				if (color_index <  10) leading_spaces++;
				fwrite(semmi, 1, leading_spaces, fp);
				if (leading_spaces < fieldlen - 2) fwrite(&(data1 = '0' + (color_index / 100)),       1, 1, fp);
				if (leading_spaces < fieldlen - 1) fwrite(&(data1 = '0' + ((color_index / 10) % 10)), 1, 1, fp);
				fwrite(&(data1 = '0' + (color_index % 10)), 1, 1, fp);
			}
		}
		image += stride - width;
	}
	fwrite(&(data1 = 0x1A), 1,  1, fp);  // file terminator
	fclose(fp);
}


int ExportPointsToSHPX(string filename, byte* image, int stride, int width, int height, int pixelsize)
{
	// declarations, initializations
	int                 val4;
	double				val8;
	FILE				*f_shp;
	FILE				*f_shx;
	string				file_shp = filename + ".shp";
	string				file_shx = filename + ".shx";
	shp_file_header		shp_fh(0, width, height);
	int					rec_count = 0;

	// open file
	if ( (f_shp = fopen(file_shp.c_str(),"wb")) == 0) return -1;

	fwrite(&shp_fh, sizeof(shp_file_header),   1, f_shp);

	// write out points
	if (pixelsize == 8)
	{
		for (int y = 0; y < height; y++)
		for (int x = 0; x < width; x++)
		if (image[stride * y + x])
		{
			++rec_count;
			// record header
			fwrite(&(val4 = LE_TO_BE_INT(rec_count)), 4, 1, f_shp); // record number
			fwrite(&(val4 = LE_TO_BE_INT(10)),        4, 1, f_shp); // content length
			// record
			fwrite(&(val4 = 1),                       4, 1, f_shp); // shape type
			fwrite(&(val8 = x),                       8, 1, f_shp);
			fwrite(&(val8 = height - 1 - y),          8, 1, f_shp);
		}
	}
	else if (pixelsize == 1)
	{
		for (int y = 0; y < height; y++)
		for (int x = 0; x < width; x++)
		if ( (image[stride * y + (x >> 3)] >> (~x & 7)) & 1 )
		{
			++rec_count;
			// record header
			fwrite(&(val4 = LE_TO_BE_INT(rec_count)), 4, 1, f_shp); // record number
			fwrite(&(val4 = LE_TO_BE_INT(10)),        4, 1, f_shp); // content length
			// record
			fwrite(&(val4 = 1),                       4, 1, f_shp); // shape type
			fwrite(&(val8 = x),                       8, 1, f_shp);
			fwrite(&(val8 = height - 1 - y),          8, 1, f_shp);
		}
	}

	// finalize
	fseek(f_shp, 0, SEEK_SET);
	shp_fh.FileLength = 50 + 14 * rec_count;
	shp_fh.FileLength = LE_TO_BE_INT(shp_fh.FileLength);
	fwrite(&shp_fh, sizeof(shp_file_header),   1, f_shp);
	fclose(f_shp);


	// ----------------------

	// open index file
	if ( (f_shx = fopen(file_shx.c_str(),"wb")) == 0) return -1;

	shp_fh.FileLength = 50 + 4 * rec_count;
	shp_fh.FileLength = LE_TO_BE_INT(shp_fh.FileLength);
	fwrite(&shp_fh, sizeof(shp_file_header),   1, f_shp);

	// values measured in 16bit
	for (int i = 1; i <= rec_count; i++)
	{
		fwrite(&(val4 = LE_TO_BE_INT((i-1)*14 + 50)), 4, 1, f_shx); // offset
		fwrite(&(val4 = LE_TO_BE_INT(10)),            4, 1, f_shx); // content length
	}
	fclose(f_shx);

	return rec_count;
}





// ---------------------------------------------------------------------------------------------------------------

namespace ImageSaver 
{
	bool ImageSaver::DumpErrorMessage(String^ msg, xml_node^ xn)
	{
		MessageBox::Show("SHP Points modul: " + msg + " fail");
		delete xn;
		return false;
	}

	String^ ImageSaver::GetExportModuleName()
	{
		return "SHP Points (from 1bpp and 8bpp image)";
	}

	int ImageSaver::GetAPIVersion()
	{
		return IMAGEFILTER_API_VERSION;
	}

	bool ImageSaver::Save_Image(String^ filename, IntPtr ptr, IntPtr size, IntPtr xml)
	{
		xml_node^ root = gcnew xml_node( xml.ToPointer() );
		xml_node^ alloc;
		int width, height, stride, packed, interleaved, channelcount, pixelsize;
		int *img_ptr  = (int*)ptr.ToPointer();
		int *img_size = (int*)size.ToPointer();
		Byte* data;

		if (FilterSpace::ImageFilterBase::GetOneImageSetting(root, width, height, stride, packed, interleaved, channelcount, pixelsize, data))
			return DumpErrorMessage("GetOneImageSetting", root);

		if (channelcount != 1) return DumpErrorMessage("A kp 1 csatorns kell legyen!", root);
		if (! root->XmlGetChildByName(alloc, "ChannelsBitAllocation")) return DumpErrorMessage("A kp csatornalersa nem tallhat!", root);
		if (pixelsize != 8 && pixelsize != 1) return DumpErrorMessage("A kp csatornja 1 vagy 8 bites kell legyen!", root);

		// --------------------------------------------
		// Framework guarantees, that stride is a dword aligned size

		string str_filename = StringTostring(filename);
		int    point_count  = ExportPointsToSHPX(str_filename, data, stride, width, height, pixelsize);

		if (point_count < 0)
		{
			delete root;
			MessageBox::Show("Unable to Save SHP/SHX files - ExportPointsToSHPX Error");
			return false;
		}
		ExportPointsToDBF( str_filename, data, stride, width, height, point_count);

		delete root;
		return true;
	}
}